package cz.uhk.iot.service;

import com.fazecast.jSerialComm.SerialPort;
import com.fazecast.jSerialComm.SerialPortDataListener;
import com.fazecast.jSerialComm.SerialPortEvent;
import cz.uhk.iot.gui.MainFrame;
import cz.uhk.iot.model.AuthenticationRequest;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.*;
import java.util.stream.Collectors;

@Service
public class SerialCommunicationService {

    private MainFrame window;

    private SerialPort[] ports;
    private Map<String, SerialPort> portMap = new HashMap();

    private SerialPort serialPort;
    private InputStream inputStream;
    private OutputStream outputStream;

    private boolean bConnected = false;

    private Logger LOG = LoggerFactory.getLogger(SerialCommunicationService.class);

    private final RadiusService radiusService;
    private final SecurityService securityService;

    @Autowired
    public SerialCommunicationService(RadiusService radiusService, SecurityService securityService) {
        this.radiusService = radiusService;
        this.securityService = securityService;
    }

    public boolean init(MainFrame window) {
        try {
            this.window = window;
            searchForPorts();
            return true;
        } catch (Exception e) {
            LOG.error("Failed to initialize {}", this.getClass(), e);
            return false;
        }
    }

    private void searchForPorts() {
        ports = SerialPort.getCommPorts();
        for (SerialPort curPort : ports) {
            portMap.put(curPort.getSystemPortName(), curPort);
        }
    }

    public void connect(String portName) {
        serialPort = portMap.get(portName);
        serialPort.setComPortParameters(9600, 8, 1, 0);
        serialPort.setComPortTimeouts(SerialPort.TIMEOUT_WRITE_BLOCKING, 0, 0);

        try {
            serialPort.openPort();
            inputStream = serialPort.getInputStream();
            outputStream = serialPort.getOutputStream();
            initPortListener();
            setConnected(true);
        } catch (Exception e) {
            LOG.error("Failed to open {}", portName, e);
        }
    }

    private void initPortListener() {
        serialPort.addDataListener(new SerialPortDataListener() {
            private String messages = "";

            @Override
            public int getListeningEvents() {
                return SerialPort.LISTENING_EVENT_DATA_RECEIVED;
            }

            @Override
            public void serialEvent(SerialPortEvent event) {
                try {
                    Thread.sleep(100); //wait for buffer
                } catch (InterruptedException e) {
                    LOG.error("Thread was interrupted", e);
                }
                messages += new String(event.getReceivedData());
                while (messages.contains("\n")) {
                    String[] message = messages.split("\\r\\n", 2);
                    messages = (message.length > 1) ? message[1] : "";
                    receiveData(message[0]);
                }
            }
        });
    }

    public void disconnect() {
        try {
            serialPort.removeDataListener();
            serialPort.closePort();
            inputStream.close();
            outputStream.close();
            setConnected(false);

            LOG.info("Disconnect was successful");
        } catch (Exception e) {
            LOG.error("Failed to close {}", serialPort.getSystemPortName(), e);
        }
    }

    final public boolean getConnected() {
        return bConnected;
    }

    private void setConnected(boolean bConnected) {
        this.bConnected = bConnected;
    }


    public void writeData(String data) {
        try {
            OutputStreamWriter osWriter = new OutputStreamWriter(outputStream);
            osWriter.write(data);
            osWriter.flush();
            LOG.info("Writing data '{}'", data);
        } catch (Exception e) {
            LOG.error("Failed to write data", e);
        }
    }

    public List<String> getPortsNames() {
        List<String> portNames = new ArrayList<>();
        portNames.addAll(Arrays.stream(ports).map(SerialPort::getSystemPortName).collect(Collectors.toList()));
        return portNames;
    }

    private void receiveData(String data) {
        if (data != null) {
            LOG.debug(data);
            if (data.contains("auth")) {
                handleAuthenticationRequestFromNode(data);
            } else if (data.contains("temp")) {
                handleTemperature(data);
            } else {
                handleStatusData(data);
            }
        }
    }

    private void handleStatusData(String data) {
        String[] splitData = data.split("_");
        StringBuilder stringBuilder = new StringBuilder();
        for (int i = 0; i < splitData.length; i++) {
            stringBuilder.append(splitData[i]);
            if (i != splitData.length - 1) {
                stringBuilder.append(": ");
            }
        }
        window.setStatusData(stringBuilder.toString());
    }

    private void handleTemperature(String data) {
        window.setTemperature(data);
    }

    private void handleAuthenticationRequestFromNode(String input) {
        AuthenticationRequest cr = new AuthenticationRequest(input);

        if (radiusService.authenticate(cr.getUsername(), securityService.generateRadiusPassFromHash(cr.getPassword()))) {
            String authResponse = "";
            if (cr.getNodeType().equals("BT")) {
                authResponse += "bA";
            } else {
                authResponse += "xA";
            }
            authResponse += securityService.generateAuthAcceptToken(cr.getUsername(), cr.getCrcNum());
            writeData(authResponse);
            window.setStatusData("Uzel " + cr.getUsername() + " byl autentizován");
        }
    }
}
